/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <string.h>

//#define DEBUG_ENABLED
#define WIN32

#include "config.h"
#ifdef HAVE_LIBMMX
#  include "mmx.h"
#endif

#include "gstwindec.h"
#include <libs/riff/gstriff.h>

#include <wine/driver.h>
#include <wine/vfw.h>

/* elementfactory information */
GstElementDetails gst_windec_details = {
  "Windows codec decoder",
  "Filter/Decoder/Image",
  NULL,
  VERSION,
  "Wim Taymans <wim.taymans@chello.be>",
  "(C) 2000",
};

extern GstPadTemplate *wincodec_src_temp;
extern GstPadTemplate *wincodec_sink_temp;

/* Winec signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_AVI_HEADER,
  ARG_BITDEPTH,
  /* FILL ME */
};

static void 	gst_windec_class_init		(GstWinDec *klass);
static void 	gst_windec_init			(GstWinDec *windec);

static void 	gst_windec_chain 		(GstPad *pad, GstBuffer *buf);

static void 	gst_windec_set_arg		(GtkObject *object,GtkArg *arg,guint id);
static void 	gst_windec_get_arg		(GtkObject *object,GtkArg *arg,guint id);

static int formats[] = { 32, 24, 16, 15, 0 };

typedef struct _GstWinLoaderData GstWinLoaderData;

struct _GstWinLoaderData {
  COMPVARS m_cv;
  BITMAPINFOHEADER m_obh;
  BITMAPINFOHEADER m_bh;
  guchar uncompr[1280000];
};


static GstElementClass *parent_class = NULL;
//static guint gst_windec_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_windec_get_type (void) 
{
  static GtkType windec_type = 0;

  if (!windec_type) {
    static const GtkTypeInfo windec_info = {
      "GstWinDec",
      sizeof(GstWinDec),
      sizeof(GstWinDec),
      (GtkClassInitFunc)gst_windec_class_init,
      (GtkObjectInitFunc)gst_windec_init,
      (GtkArgSetFunc)gst_windec_set_arg,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    windec_type = gtk_type_unique(GST_TYPE_ELEMENT,&windec_info);
  }
  return windec_type;
}

static void
gst_windec_class_init (GstWinDec *klass) 
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtk_object_add_arg_type("GstWinDec::avi_header", GTK_TYPE_POINTER,
                               GTK_ARG_WRITABLE, ARG_AVI_HEADER);
  gtk_object_add_arg_type("GstWinDec::bitdepth", GTK_TYPE_INT,
                               GTK_ARG_READWRITE, ARG_BITDEPTH);

  gtkobject_class->set_arg = gst_windec_set_arg;
  gtkobject_class->get_arg = gst_windec_get_arg;
}

static GstCaps* 
gst_windec_bh_to_caps (BITMAPINFOHEADER *bh) 
{
    GstCaps *caps;
    gulong compression;
    gint bitCount;
    gint red_mask, green_mask, blue_mask;

    if (bh->biCompression == 0) 
      compression = GST_MAKE_FOURCC ('R','G','B',' ');
    else
      compression = bh->biCompression;

    if (compression != GST_MAKE_FOURCC ('R','G','B',' ')) {
      caps = gst_caps_new (
		    "windec_caps",
		    "video/raw",
		    gst_props_new (
			  "format",    GST_PROPS_FOURCC (compression),
	                    "width",      GST_PROPS_INT (bh->biWidth),
	                    "height",     GST_PROPS_INT (bh->biHeight),
			    NULL));
    }
    else {
      bitCount = bh->biBitCount;

      switch (bitCount) {
        case 15:
          red_mask   = 0x7c00;
          green_mask = 0x03e0;
          blue_mask  = 0x001f;
	  break;
        case 16:
          red_mask   = 0xf800;
          green_mask = 0x07e0;
          blue_mask  = 0x001f;
	  break;
        case 32:
        case 24:
          red_mask   = 0x00ff0000;
          green_mask = 0x0000ff00;
          blue_mask  = 0x000000ff;
	  break;
        default:
	  g_warning ("gstwindec: unknown bitCount found\n");
	  red_mask = green_mask = blue_mask = 0;
	  break;
      }

      caps = gst_caps_new (
		    "windec_caps",
		    "video/raw",
		    gst_props_new (
			  "format",    GST_PROPS_FOURCC (compression),
	                    "bpp",        GST_PROPS_INT (bitCount),
	                    "depth",      GST_PROPS_INT (bitCount),
			    "endianness", GST_PROPS_INT (G_LITTLE_ENDIAN),
			    "red_mask",   GST_PROPS_INT (red_mask),
			    "green_mask", GST_PROPS_INT (green_mask),
			    "blue_mask",  GST_PROPS_INT (blue_mask),
	                    "width",      GST_PROPS_INT (bh->biWidth),
	                    "height",     GST_PROPS_INT (bh->biHeight),
			    NULL));
    }
    return caps;
}

static GstPadNegotiateReturn 
gst_windec_negotiate_src (GstPad *pad, GstCaps **caps, gpointer *user_data) 
{
  GstWinDec *windec;
  GstWinLoaderData *data;
  HRESULT h;
  BITMAPINFOHEADER temp;
  static int hack_counter = 0; //FIXME

  windec = GST_WINDEC (gst_pad_get_parent (pad));

  data = (GstWinLoaderData *)windec->extra_data;
  GST_INFO (GST_CAT_NEGOTIATION, "gstwindec: src negotiate\n");

  memcpy(&temp, &data->m_obh, sizeof(data->m_obh));

  if (*caps) {
    temp.biBitCount = gst_caps_get_int (*caps, "depth");
    temp.biCompression = gst_caps_get_int (*caps, "format");

    GST_INFO (GST_CAT_NEGOTIATION, "got %08lx\n", temp.biCompression);

    if (temp.biCompression == GST_MAKE_FOURCC ('Y','U','Y','2') ||
        temp.biCompression == GST_MAKE_FOURCC ('R','G','B',' ')) 
    {
      if (temp.biCompression == GST_MAKE_FOURCC ('R','G','B',' ')) {
        temp.biSizeImage = data->m_obh.biWidth * data->m_obh.biHeight * 
	        gst_caps_get_int (*caps, "bpp");
      }

      h = ICDecompressQuery(data->m_cv.hic, &data->m_bh, &temp);
      if (h == 0) {
        memcpy(&data->m_obh, &temp, sizeof(data->m_obh));
        GST_INFO(GST_CAT_NEGOTIATION, "gstwindec:: Set Bit Depth to %d\n", temp.biBitCount);

        return GST_PAD_NEGOTIATE_AGREE;
      }
      else {
        GST_INFO(GST_CAT_NEGOTIATION, "gstwindec:: Set Bit Depth failed\n");
      }
    }
    *caps = NULL;
    return GST_PAD_NEGOTIATE_TRY;
  }
  else {

    temp.biBitCount = formats[hack_counter++];
    if (temp.biBitCount == 0)
      return GST_PAD_NEGOTIATE_FAIL;

    temp.biSizeImage = data->m_obh.biWidth * data->m_obh.biHeight * 
                            temp.biBitCount;

    h = ICDecompressQuery(data->m_cv.hic, &data->m_bh, &temp);

    if (h == 0) {
      memcpy (&data->m_obh, &temp, sizeof(data->m_obh));
      GST_INFO (GST_CAT_NEGOTIATION, "gstwindec: Suggest bit depth %d\n", data->m_obh.biBitCount);
      *caps = gst_windec_bh_to_caps (&data->m_obh);
    }
    else {
      GST_INFO(GST_CAT_NEGOTIATION, "gstwindec: Set Bit Depth failed\n");
      *caps = NULL;
    }
    return GST_PAD_NEGOTIATE_TRY;
  }

  return GST_PAD_NEGOTIATE_FAIL;
}

static void 
gst_windec_newcaps (GstPad *pad, GstCaps *newcaps) 
{
  GstWinDec *windec;
  GstWinLoaderData *data;

  windec = GST_WINDEC (gst_pad_get_parent (pad));

  data = (GstWinLoaderData *)windec->extra_data;

  data->m_bh.biSize 		= gst_caps_get_int (newcaps, "size");
  data->m_bh.biWidth 		= gst_caps_get_int (newcaps, "width");
  data->m_bh.biHeight 		= gst_caps_get_int (newcaps, "height");
  data->m_bh.biPlanes 		= gst_caps_get_int (newcaps, "planes");
  data->m_bh.biBitCount 	= gst_caps_get_int (newcaps, "bit_cnt");
  data->m_bh.biCompression 	= gst_caps_get_fourcc_int (newcaps, "compression");
  data->m_bh.biSizeImage 	= gst_caps_get_int (newcaps, "image_size");
  data->m_bh.biXPelsPerMeter 	= gst_caps_get_int (newcaps, "xpels_meter");
  data->m_bh.biYPelsPerMeter 	= gst_caps_get_int (newcaps, "ypels_meter");
  data->m_bh.biClrUsed 		= gst_caps_get_int (newcaps, "num_colors");
  data->m_bh.biClrImportant 	= gst_caps_get_int (newcaps, "imp_colors");
}

static void 
gst_windec_init (GstWinDec *windec) 
{
  GST_DEBUG (0,"gst_windec_init: initializing\n");
  /* create the sink and src pads */
  windec->sinkpad = gst_pad_new_from_template (wincodec_sink_temp, "sink");
  gst_element_add_pad (GST_ELEMENT (windec), windec->sinkpad);
  gst_pad_set_chain_function (windec->sinkpad, gst_windec_chain);
  gst_pad_set_newcaps_function (windec->sinkpad, gst_windec_newcaps);

  windec->srcpad = gst_pad_new_from_template (wincodec_src_temp, "src");
  gst_element_add_pad (GST_ELEMENT (windec), windec->srcpad);
  gst_pad_set_negotiate_function (windec->srcpad, gst_windec_negotiate_src);

  /* initialize the windec decoder state */
  windec->next_time = 0;

  // reset the initial video state
  windec->width = -1;
  windec->height = -1;

  windec->inited = FALSE;

  windec->extra_data = g_malloc0 (sizeof (GstWinLoaderData));
}

static gboolean 
gst_windec_init_decoder (GstWinDec *windec)
{
  HRESULT h;
  GstWinLoaderData *data;

  data = (GstWinLoaderData *)windec->extra_data;

  data->m_obh.biSize=sizeof(data->m_obh);
  data->m_cv.cbSize=sizeof(data->m_cv);
  GST_INFO (GST_CAT_PLUGIN_INFO, "gstwindec: trying to open library %p\n", data);
  data->m_cv.hic=ICOpen(
           GST_RIFF_FCC_vidc,
           data->m_bh.biCompression,
           ICMODE_DECOMPRESS);
  if(data->m_cv.hic==0)
  {
     printf("gstwindec:: ICOpen failed\n");
     return FALSE;
  }
  data->m_cv.fccType=GST_RIFF_FCC_vidc;
  data->m_cv.fccHandler=data->m_bh.biCompression,


  h=ICDecompressGetFormat(data->m_cv.hic, &data->m_bh, &data->m_obh);
  if(h!=0)
  {
    printf("gstwindec:: Can't handle this format\n");
    ICClose(data->m_cv.hic);
    return FALSE;
  }

  data->m_obh.biBitCount=16;
  data->m_obh.biCompression = GST_MAKE_FOURCC ('Y','U','Y','2');

  h = ICDecompressQuery(data->m_cv.hic, &data->m_bh, &data->m_obh);

  if (!gst_pad_set_caps (windec->srcpad, gst_windec_bh_to_caps (&data->m_obh))) {
    return FALSE;
  }
  
  GST_INFO (GST_CAT_PLUGIN_INFO, "gstwindec:: output format %08lx (%4.4s) %ld\n", data->m_obh.biCompression, 
		  (gchar *)&data->m_obh.biCompression, data->m_obh.biHeight);

  h=ICDecompressBegin(data->m_cv.hic, &data->m_bh, &data->m_obh);
  if(h!=0)
  {
    printf("gstwindec:: ICDecompressBegin failed\n");
    return FALSE;
  }

  return TRUE;
}

#ifdef HAVE_LIBMMX
static mmx_t MMX_FFE0           = (mmx_t)(long long)0xFFE0FFE0FFE0FFE0;
#endif

static void 
gst_windec_chain (GstPad *pad, GstBuffer *buf)
{
  GstWinDec *windec;
  GstWinLoaderData *data;
  GstBuffer *outbuf;
  guint32 *outdata;
  guint32 *ptr, *outptr;
  gint i, j;
  gint w;

  g_return_if_fail(buf != NULL);
  if (GST_BUFFER_DATA(buf) == NULL) return;

  windec = GST_WINDEC (gst_pad_get_parent (pad));

  if (!GST_PAD_CONNECTED(windec->srcpad)) {
    GST_DEBUG (0,"gst_windec: src pad not connected\n");
    gst_buffer_unref(buf);
    return;
  }

  if (!windec->inited) {
    if (!gst_windec_init_decoder(windec)) {
      g_print("gst_windec: could not init decoder\n");
      return;
    }
    windec->inited = TRUE;
  }

  data = (GstWinLoaderData *)windec->extra_data;

  outbuf = NULL;
  if (!windec->pool) {
    windec->pool = gst_pad_get_bufferpool (windec->srcpad);
  }
  if (windec->pool) {
    outbuf = gst_buffer_new_from_pool (windec->pool);
  }

  if (!outbuf) {
    outbuf = gst_buffer_new ();
    GST_BUFFER_SIZE (outbuf) = data->m_obh.biWidth *
  			       data->m_obh.biHeight *
  			       ((data->m_obh.biBitCount+7)/8);
    GST_BUFFER_DATA (outbuf) = g_malloc (GST_BUFFER_SIZE (outbuf));
  }

  outdata = (guint32 *)GST_BUFFER_DATA (outbuf);

  GST_DEBUG (0,"gst_windec: got buffer %08lx %p %u\n", *(gulong *)GST_BUFFER_DATA (buf), data, GST_BUFFER_SIZE (buf));
  
  ICDecompress(data->m_cv.hic, 
		  0&ICDECOMPRESS_NOTKEYFRAME, 
		  &data->m_bh, 
		  GST_BUFFER_DATA (buf), 
		  &data->m_obh, 
		  &data->uncompr[0]);

  GST_DEBUG (0, "gst_windec: decompressed buffer %d %ld %ld\n",
             data->m_obh.biBitCount, data->m_obh.biWidth,
             data->m_obh.biHeight);

  GST_BUFFER_TIMESTAMP (outbuf) = GST_BUFFER_TIMESTAMP (buf);
  if (GST_BUFFER_FLAG_IS_SET (buf, GST_BUFFER_FLUSH))
    GST_BUFFER_FLAG_SET (outbuf, GST_BUFFER_FLUSH);

  ptr = (guint32 *) &data->uncompr[0];
  w = (data->m_obh.biWidth * ((data->m_obh.biBitCount+7)/8))/2;
  outptr = (guint32 *) (((gchar*)outdata) + (GST_BUFFER_SIZE (outbuf) - w*2));

  GST_DEBUG (0, "%d %d %d %ld %d\n", w, (gchar*)outptr - (gchar*)outdata, 640*480*2, 
		  data->m_obh.biSizeImage, GST_BUFFER_SIZE (outbuf));

  /* avi video is upside down so we need to flip it.
   * avi 16 bits actually is only 15 bits so we need to convert that too */
  if (data->m_obh.biBitCount == 16 &&
      data->m_obh.biCompression != GST_MAKE_FOURCC ('Y','U','Y','2')) {
#ifdef HAVE_LIBMMX
    movq_m2r(MMX_FFE0, mm3);
#endif
    for (i=data->m_obh.biHeight; i ; i--) {
      for (j=(w>>2); j ; j--) {
#ifdef HAVE_LIBMMX
        movq_m2r(*ptr, mm0);
        movq_r2r(mm0, mm2);
        pand_r2r(mm3, mm2);
        paddw_r2r(mm2, mm0);
        movq_r2m(mm0, *outptr);
        ptr+=2;
        outptr+=2;
#else
        *outptr++ = *ptr + (*ptr++&0xFFE0FFE0);
        *outptr++ = *ptr + (*ptr++&0xFFE0FFE0);
#endif
      }
      outptr -= w;
    }
  }
  else {
    for (i=data->m_obh.biHeight; i ; i--) {
      for (j=(w>>2); j ; j--) {
#ifdef HAVE_LIBMMX
        movq_m2r(*ptr, mm0);
        movq_r2m(mm0, *outptr);
        ptr+=2;
        outptr+=2;
#else
        *outptr++ = ptr++;
        *outptr++ = ptr++;
#endif
      }
      outptr -= w;
    }
  }
#ifdef HAVE_LIBMMX
  emms();
#endif


  GST_DEBUG (0,"gst_windec: pushing buffer %llu\n", GST_BUFFER_TIMESTAMP (outbuf));
  gst_pad_push (windec->srcpad, outbuf);
  GST_DEBUG (0,"gst_windec: pushed buffer\n");

  gst_buffer_unref (buf);
}

static void 
gst_windec_set_arg (GtkObject *object, GtkArg *arg, guint id)
{
  GstWinDec *windec;
  GstWinLoaderData *data;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_WINDEC(object));
  windec = GST_WINDEC(object);

  data = (GstWinLoaderData *)windec->extra_data;

  switch(id) {
    case ARG_AVI_HEADER:
      memcpy(&data->m_bh, GTK_VALUE_POINTER(*arg), sizeof(BITMAPINFOHEADER));
      break;
    case ARG_BITDEPTH:
      break;
    default:
      break;
  }
}

static void 
gst_windec_get_arg (GtkObject *object, GtkArg *arg, guint id)
{
  GstWinDec *windec;
  GstWinLoaderData *data;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_WINDEC(object));
  windec = GST_WINDEC(object);

  data = (GstWinLoaderData *)windec->extra_data;

  switch(id) {
    case ARG_BITDEPTH:
      break;
    default:
      break;
  }
}

