/* Gnome-Streamer
 * Copyright (C) <2001> Steve Baker <stevebaker_org@yahoo.co.uk>
 *
 * int2float.c
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <int2float.h>


static GstElementDetails int2float_details = {
  "Integer to Float effect",
  "Filter/Effect",
  "Convert from integer to floating point audio data",
  VERSION,
  "Steve Baker <stevebaker_org@yahoo.co.uk>",
  "(C) 2001",
};

/* Stereo signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_PAN,
};

static GstPadTemplate* 
int2float_src_factory (void)
{
  return 
    gst_padtemplate_new (
    "src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    gst_caps_new (
    "int2float_src",
      "audio/raw",
      gst_props_new (
        "format",     GST_PROPS_STRING ("float"),
        "layout",     GST_PROPS_STRING ("gfloat"),
        "intercept",  GST_PROPS_FLOAT(0.0),
        "slope",      GST_PROPS_FLOAT(1.0),
        "channels",   GST_PROPS_INT (1),
      NULL)),
    NULL);
}

static GstPadTemplate* 
int2float_sink_factory (void)
{
  return 
    gst_padtemplate_new (
    "sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    gst_caps_new (
    "int2float_sink",
      "audio/raw",
      gst_props_new (
        "format",   GST_PROPS_STRING ("int"),
        "law",     GST_PROPS_INT (0),
        "endianness",     GST_PROPS_INT (G_BYTE_ORDER),
        "signed",   GST_PROPS_BOOLEAN (TRUE),
        "channels", GST_PROPS_INT (1),
        "width", GST_PROPS_INT (16),
        "depth", GST_PROPS_INT (16),
      NULL)),
    NULL);
}

static GstPadTemplate *srctempl, *sinktempl;

static void	gst_int2float_class_init		(GstInt2FloatClass *klass);
static void	gst_int2float_init			(GstInt2Float *plugin);
static void	gst_int2float_set_arg		(GtkObject *object, GtkArg *arg, guint id);
static void	gst_int2float_get_arg		(GtkObject *object, GtkArg *arg, guint id);
static void	gst_int2float_chain_gint16  (GstPad *pad, GstBuffer *buf_in);
static inline GstInt2Float* gst_int2float_get_plugin(GstPad *pad,GstBuffer *buf);

static GstElementClass *parent_class = NULL;

static GstPadNegotiateReturn
gst_int2float_negotiate_sink (GstPad *pad, GstCaps **caps, gpointer *data)
{
  GstInt2Float* plugin = GST_INT2FLOAT (GST_OBJECT_PARENT (pad));
  GstCaps* peercaps;
  GstRealPad *realpeer;
  GstPadNegotiateReturn proxyresult;
  
  g_print("nego int2float sink\n");
  if (*caps== NULL) return GST_PAD_NEGOTIATE_FAIL;
  
  plugin->width = gst_caps_get_int (*caps, "width");
  plugin->depth = gst_caps_get_int (*caps, "depth");
  
  if (plugin->trying_proxy_caps) return GST_PAD_NEGOTIATE_AGREE;

  plugin->trying_proxy_caps=TRUE;
  
  peercaps=gst_caps_new (
    "int2float_sink_caps",
    "audio/raw",
    gst_props_new (
      "format",     GST_PROPS_STRING ("float"),
      "layout",     GST_PROPS_STRING ("gfloat"),
      "intercept",  GST_PROPS_FLOAT(0.0),
      "slope",      GST_PROPS_FLOAT(1.0),
      "rate",       GST_PROPS_INT (gst_caps_get_int (*caps, "rate")),
      "channels",   GST_PROPS_INT (1),
      NULL
    )
  );
  realpeer = GST_RPAD_PEER(plugin->srcpad);
  proxyresult = gst_pad_negotiate_proxy(&(realpeer->pad), plugin->srcpad, &peercaps);
  plugin->trying_proxy_caps=FALSE;
  return proxyresult;
}

static GstPadNegotiateReturn
gst_int2float_negotiate_src (GstPad *pad, GstCaps **caps, gpointer *data)
{
  GstInt2Float* plugin = GST_INT2FLOAT (GST_OBJECT_PARENT (pad));
  GstCaps* peercaps;
  GstRealPad *realpeer;
  GstPadNegotiateReturn proxyresult;

  g_print("nego int2float src\n");
  if (*caps== NULL) return GST_PAD_NEGOTIATE_FAIL;

  if (plugin->trying_proxy_caps) return GST_PAD_NEGOTIATE_AGREE;

  plugin->trying_proxy_caps=TRUE;
  peercaps=gst_caps_new (
    "int2float_src_caps",
    "audio/raw",
    gst_props_new (
      "format",     GST_PROPS_STRING ("int"),
      "law",        GST_PROPS_INT (0),
      "endianness", GST_PROPS_INT (G_BYTE_ORDER),
      "signed",     GST_PROPS_BOOLEAN (TRUE),
      "width",      GST_PROPS_INT (plugin->width),
      "depth",      GST_PROPS_INT (plugin->depth),
      "rate",       GST_PROPS_INT (gst_caps_get_int (*caps, "rate")),
      "channels",   GST_PROPS_INT (1),
      NULL
    )
  );
  realpeer = GST_RPAD_PEER(plugin->sinkpad);
  proxyresult = gst_pad_negotiate_proxy(&(realpeer->pad), plugin->sinkpad, &peercaps);
  plugin->trying_proxy_caps=FALSE;
  return proxyresult;
}

GtkType
gst_int2float_get_type(void) {
  static GtkType int2float_type = 0;

  if (!int2float_type) {
    static const GtkTypeInfo int2float_info = {
      "GstInt2Float",
      sizeof(GstInt2Float),
      sizeof(GstInt2FloatClass),
      (GtkClassInitFunc)gst_int2float_class_init,
      (GtkObjectInitFunc)gst_int2float_init,
      (GtkArgSetFunc)gst_int2float_set_arg,
      (GtkArgGetFunc)gst_int2float_get_arg,
      (GtkClassInitFunc)NULL,
    };
    int2float_type = gtk_type_unique(GST_TYPE_ELEMENT,&int2float_info);
  }
  return int2float_type;
}

static void
gst_int2float_class_init (GstInt2FloatClass *klass)
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gtkobject_class->set_arg = gst_int2float_set_arg;
  gtkobject_class->get_arg = gst_int2float_get_arg;
}

static void
gst_int2float_init (GstInt2Float *plugin)
{
  plugin->sinkpad = gst_pad_new_from_template(sinktempl,"sink");
  plugin->srcpad = gst_pad_new_from_template(srctempl,"src");

  gst_element_add_pad(GST_ELEMENT(plugin),plugin->sinkpad);
  gst_pad_set_chain_function(plugin->sinkpad,gst_int2float_chain_gint16);
  gst_element_add_pad(GST_ELEMENT(plugin),plugin->srcpad);

  gst_pad_set_negotiate_function(plugin->sinkpad, gst_int2float_negotiate_sink);
  gst_pad_set_negotiate_function(plugin->srcpad, gst_int2float_negotiate_src);
  plugin->channels=1;
  plugin->width=16;
  plugin->depth=16;
  plugin->intercept=0.0;
  plugin->slope=1.0;
  plugin->trying_proxy_caps=FALSE;
}

static void
gst_int2float_chain_gint16 (GstPad *pad,GstBuffer *buf_in)
{
  GstInt2Float *plugin;
  GstBuffer *buf_out;
  gint16 *data_in;
  gfloat *data_out;
  gint num_samples, i;
  
  g_return_if_fail(plugin = gst_int2float_get_plugin(pad,buf_in));

  num_samples = GST_BUFFER_SIZE(buf_in)/2;
  data_in=(gint16 *)GST_BUFFER_DATA(buf_in);

  buf_out = gst_buffer_new();
  g_return_if_fail (buf_out);
  data_out = g_new(gfloat, num_samples);
  GST_BUFFER_DATA(buf_out) = (gpointer) data_out;
  GST_BUFFER_SIZE(buf_out) = sizeof(gfloat) * num_samples;
  
  for (i = 0; i < num_samples; i++) {
    data_out[i] = ((gfloat)data_in[i]) / 32767.0;
  }
  
  gst_buffer_unref(buf_in);  
  gst_pad_push(plugin->srcpad,buf_out);
}

static inline GstInt2Float* 
gst_int2float_get_plugin(GstPad *pad,GstBuffer *buf)
{
  GstInt2Float *plugin;
  g_return_val_if_fail(pad != NULL, NULL);
  g_return_val_if_fail(GST_IS_PAD(pad), NULL);
  g_return_val_if_fail(buf != NULL, NULL);

  plugin = GST_INT2FLOAT(GST_OBJECT_PARENT (pad));
  g_return_val_if_fail(plugin != NULL, NULL);
  g_return_val_if_fail(GST_IS_INT2FLOAT(plugin), NULL);
  return plugin;
}

static void
gst_int2float_set_arg (GtkObject *object, GtkArg *arg, guint id)
{
  GstInt2Float *plugin;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_INT2FLOAT(object));
  plugin = GST_INT2FLOAT(object);

  switch(id) {
    default:
      break;
  }
}

static void
gst_int2float_get_arg (GtkObject *object, GtkArg *arg, guint id)
{
  GstInt2Float *plugin;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_INT2FLOAT(object));
  plugin = GST_INT2FLOAT(object);

  switch (id) {
    default:
      arg->type = GTK_TYPE_INVALID;
      break;
  }
}

gboolean 
gst_int2float_factory_init (GstPlugin *plugin) 
{
  GstElementFactory *factory;
  
  factory = gst_elementfactory_new("int2float",GST_TYPE_INT2FLOAT,
                                   &int2float_details);
  g_return_val_if_fail(factory != NULL, FALSE);
  
  sinktempl = int2float_sink_factory();
  gst_elementfactory_add_padtemplate (factory, sinktempl);

  srctempl = int2float_src_factory();
  gst_elementfactory_add_padtemplate (factory, srctempl);
  
  gst_plugin_add_factory(plugin,factory);
  
  return TRUE;
}
