/* Gnome-Streamer
 * Copyright (C) <1999> Erik Walthinsen <omega@cse.ogi.edu>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <string.h>

#include "gstgsmenc.h"

/* elementfactory information */
GstElementDetails gst_gsmenc_details = {
  "gsm audio encoder",
  "Filter/Encoder/Audio",
  ".gsm",
  VERSION,
  "Wim Taymans <wim.taymans@chello.be>",
  "(C) 2000",
};

/* GSMEnc signals and args */
enum {
  FRAME_ENCODED,
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  /* FILL ME */
};

static void		gst_gsmenc_class_init	(GstGSMEnc *klass);
static void		gst_gsmenc_init		(GstGSMEnc *gsmenc);

static void		gst_gsmenc_chain	(GstPad *pad,GstBuffer *buf);

static GstElementClass *parent_class = NULL;
static guint gst_gsmenc_signals[LAST_SIGNAL] = { 0 };

GtkType
gst_gsmenc_get_type (void)
{
  static GtkType gsmenc_type = 0;


  if (!gsmenc_type) {
    static const GtkTypeInfo gsmenc_info = {
      "GstGSMEnc",
      sizeof(GstGSMEnc),
      sizeof(GstGSMEncClass),
      (GtkClassInitFunc)gst_gsmenc_class_init,
      (GtkObjectInitFunc)gst_gsmenc_init,
      (GtkArgSetFunc)NULL,
      (GtkArgGetFunc)NULL,
      (GtkClassInitFunc)NULL,
    };
    gsmenc_type = gtk_type_unique(GST_TYPE_ELEMENT,&gsmenc_info);
  }
  return gsmenc_type;
}

static void
gst_gsmenc_class_init (GstGSMEnc *klass)
{
  GtkObjectClass *gtkobject_class;
  GstElementClass *gstelement_class;

  gtkobject_class = (GtkObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = gtk_type_class(GST_TYPE_ELEMENT);

  gst_gsmenc_signals[FRAME_ENCODED] =
    gtk_signal_new ("frame_encoded", GTK_RUN_LAST, gtkobject_class->type,
                    GTK_SIGNAL_OFFSET (GstGSMEncClass, frame_encoded),
                    gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);

  gtk_object_class_add_signals (gtkobject_class, gst_gsmenc_signals,
                                LAST_SIGNAL);
}


static void
gst_gsmenc_init (GstGSMEnc *gsmenc)
{
  /* create the sink and src pads */
  gsmenc->sinkpad = gst_pad_new("sink",GST_PAD_SINK);
  gst_element_add_pad(GST_ELEMENT(gsmenc),gsmenc->sinkpad);
  gst_pad_set_chain_function(gsmenc->sinkpad,gst_gsmenc_chain);

  gsmenc->srcpad = gst_pad_new("src",GST_PAD_SRC);
  gst_element_add_pad(GST_ELEMENT(gsmenc),gsmenc->srcpad);

  gsmenc->state = gsm_create ();
  gsmenc->bufsize = 0;
}

static void
gst_gsmenc_chain (GstPad *pad, GstBuffer *buf)
{
  GstGSMEnc *gsmenc;
  gsm_signal *data;
  guint size;

  g_return_if_fail (pad != NULL);
  g_return_if_fail (GST_IS_PAD (pad));
  g_return_if_fail (buf != NULL);

  gsmenc = GST_GSMENC (GST_OBJECT_PARENT (pad));
	      
  data = (gsm_signal*) GST_BUFFER_DATA (buf);
  size = GST_BUFFER_SIZE (buf) / sizeof (gsm_signal);

  if (gsmenc->bufsize && (gsmenc->bufsize + size >= 160)) {
    GstBuffer *outbuf;

    memcpy (gsmenc->buffer + gsmenc->bufsize, data, (160 - gsmenc->bufsize) * sizeof (gsm_signal));

    outbuf = gst_buffer_new ();
    GST_BUFFER_DATA (outbuf) = g_malloc (33 * sizeof (gsm_byte));
    GST_BUFFER_SIZE (outbuf) = 33 * sizeof (gsm_byte);

    gsm_encode (gsmenc->state, gsmenc->buffer, (gsm_byte *) GST_BUFFER_DATA (outbuf));

    gst_pad_push (gsmenc->srcpad, outbuf);

    size -= (160 - gsmenc->bufsize); 
    data += (160 - gsmenc->bufsize);
    gsmenc->bufsize = 0;
  }

  while (size >= 160) {
    GstBuffer *outbuf;

    outbuf = gst_buffer_new ();
    GST_BUFFER_DATA (outbuf) = g_malloc (33 * sizeof (gsm_byte));
    GST_BUFFER_SIZE (outbuf) = 33 * sizeof (gsm_byte);

    gsm_encode (gsmenc->state, data, (gsm_byte *) GST_BUFFER_DATA (outbuf));

    gst_pad_push (gsmenc->srcpad, outbuf);

    size -= 160;
    data += 160;
  }

  if (size) {
    memcpy (gsmenc->buffer + gsmenc->bufsize, data, size * sizeof (gsm_signal));
    gsmenc->bufsize = size;
  }

  gst_buffer_unref(buf);
}
